/* 
 * channel - switch channel on the DCT-2000 connected to the serial
 * port and print the new channel, or just print the channel if no
 * arguments are passed.
 * 
 * by Jim Paris <jim@jtan.com>.
 * Released under the GNU Public License (GPL) version 2.
 *
 * Heavily modifed by Dave Johnson <dave-dctremote@davej.org>
 */
 

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include "crc.h"
#include "serial.h"
#include "packet.h"
#include "debug.h"
#include "remote.h"

/* Timeouts, in msec. */
#define TIMEOUT_TINY 250	/* Waiting for extra channel status */
#define TIMEOUT_SHORT 1500	/* For most messages */
#define TIMEOUT_LONG 5000 	/* Waiting for channel to change */
#define KEY_SLEEP 200		/* How long to wait after sending keypress */
#define POWER_SLEEP 6000	/* How long to wait for box to power on */

int verbose = 0;

static int send_and_get_ack(_sPacket *p)
{
	_sPacket s;
	int ret=0;
	serial_sendpacket(p);
	if(serial_getpacket(&s,TIMEOUT_SHORT)<0) {
		debug("No response to packet\n");
		ret=-1;
	} else if(!packet_is_ack(p,&s)) {
		debug("Got response, but it's not an ACK\n");
		ret=-1;
	}
	packet_cleanup(&s);
	packet_cleanup(p);
	return ret;
}

static int initialize(void)
{
	_sPacket p;
	debug("Attempting to initialize DCT\n");

	serial_flush(TIMEOUT_TINY);

	/* Send initialize_1 and get ack */
	packet_build_initialize_1(&p);
	if(send_and_get_ack(&p)<0)
		return -1;

	packet_build_initialize_2(&p);
	if(send_and_get_ack(&p)<0)
		return -1;

	/* Now the DCT should send us a status message, so wait
	   for it (and automatically acknowledge it) */
	if(serial_getpacket(&p,TIMEOUT_SHORT)<0) {
		debug("Didn't get expected status message\n");
		return -1;
	}
	packet_cleanup(&p);
	return 0;
}

static int get_channel(void)
{
	int chan;
	_sPacket p;

	debug("Reading current channel\n");

	/* Send channelquery and get ack */
	packet_build_channelquery(&p);
	if(send_and_get_ack(&p)<0)
		return -1;
	
	/* Now we expect one more packet with the 
	   channel status */
	if(serial_getpacket(&p,TIMEOUT_SHORT)<0) {
		debug("Didn't get channel status message\n");
		return -1;
	}

	if((chan=packet_extract_channel(&p))<0) {
		debug("Returned packet isn't channel status!\n");
		packet_cleanup(&p);
		return -1;
	}

	packet_cleanup(&p);
	return chan;
}

static int send_keypress(int key)
{
	_sPacket p;
	struct timespec sleeptime;

	packet_build_keypress(&p,key);
	if(send_and_get_ack(&p)<0)
		return -1;

	/* After sending a keypress, we have to wait. 
	   0.1 sec is not quite enough; 0.2 sec seems to be */
	sleeptime.tv_sec = KEY_SLEEP / 1000;
	sleeptime.tv_nsec = (KEY_SLEEP % 1000) * 1000 * 1000;
	while ((nanosleep(&sleeptime, &sleeptime) < 0) && (errno == EINTR));

	return 0;
}

static int set_channel(int chan)
{
	_sPacket p;
	int cur;

	if((send_keypress(KEY_0+((chan/100)%10)))<0 ||
	   (send_keypress(KEY_0+((chan/10)%10)))<0 ||
	   (send_keypress(KEY_0+((chan/1)%10)))<0) {
		debug("error sending channel keypresses\n");
		return -1;
	}
	
	/* Now the DCT should send us channel status */
	if(serial_getpacket(&p,TIMEOUT_LONG)<0) {
		debug("didn't get channel status message\n");
		return -1;
	}
	if((cur=packet_extract_channel(&p))<0) {
		debug("Returned packet isn't channel status!\n");
		packet_cleanup(&p);
		return -1;
	}
	packet_cleanup(&p);

	/* The DCT may send more channel status messages, if it
	   figures out something new about the signal.  Wait
	   for these, but don't wait long. */
	while(serial_getpacket(&p,TIMEOUT_TINY)>=0) {
		int newcur;
		if((newcur=packet_extract_channel(&p))>=0) {
			debug("Got extra status back\n");
			cur = newcur;
		}
		packet_cleanup(&p);
	}
	packet_cleanup(&p);

	/* Send EXIT to clear the OSD, but don't
	   care if this doesn't work */
	send_keypress(KEY_EXIT);

	return cur;
}

static void usage(const char *progname)
{
	int keycode;

	fprintf(stderr,"Usage: %s [-v] [-d <device>] -c <channel>\n", progname);
	fprintf(stderr,"Usage: %s [-v] [-d <device>] -k <keys>\n", progname);

	fprintf(stderr,"  <keys> is one or more of:\n");
	for (keycode = 0; keycode < NUM_KEYS; keycode++) {
		if (key_chars[keycode])
			fprintf(stderr,"    %c for %s\n", key_chars[keycode], key_names[keycode]);
	}
	exit(-1);
}

int main(int argc, char *argv[])
{
	int retval;
	int tries;
	int cur=0;
	int chan = -1;
	const char *keys = NULL;
	const char *device = "/dev/ttyS0";

	while ( (retval = getopt(argc, argv, "vd:c:k:")) >= 0 ) {
		switch(retval) {
			case 'v':
				verbose++;
				break;

			case 'd':
				device = optarg;
				break;

			case 'c':
				chan = atoi(optarg);
				break;

			case 'k':
				keys = optarg;
				break;

			case ':':
			case '?':
			default:
				usage(argv[0]);
		}
	}

	if (optind < argc)
		usage(argv[0]);

	if (chan != -1) {
		if (keys || (chan < 1) || (chan > 999))
			usage(argv[0]);
	}
	else if (!keys)
		usage(argv[0]);

	if((serial_init(device))<0) 
		err(1,device);

	atexit(serial_close);

	if (keys) {
		if(initialize()<0) {
			debug("Initialization failed\n");
			return 1;
		}

		for (; *keys; keys++) {
			int keycode;
			for (keycode = 0; keycode < NUM_KEYS; keycode++) {
				if (key_chars[keycode] == *keys)
					break;
			}
			if (keycode == NUM_KEYS) {
				fprintf(stderr,"Unknown key %c\n", *keys);
				return 1;
			}
			debug("Sending key %c (code 0x%02x)\n", *keys, keycode);
			if (send_keypress(keycode) < 0) {
				fprintf(stderr,"Communication failed for key %c (code 0x%02x)\n", *keys, keycode);
				return 1;
			}
		}
		return 0;
	}

	for(tries = 0; tries < 4; tries++) {
		int newcur;

		if(initialize()<0) {
			debug("Initialization failed\n");
			continue;
		}

		if((newcur=get_channel())<0) {
			cur = -1;
			debug("get_channel failed\n");
                        if (!tries) {
                                continue;
                        }
		}
                else {
                        cur = newcur;
                }

		if(chan!=0 && chan!=cur) {
			/* Switch channel */
			if((newcur=set_channel(chan))<0) {
				int i;

				debug("set_channel failed\n");
				if(tries==1 || tries==3) {
					/* Try sending exit key a few times to make
					   sure we're not in a menu or something */
					debug("Sending KEY_EXIT key\n");
					for(i=0;i<4;i++)
						if(send_keypress(KEY_EXIT)<0) break;
				}
#if 0
				else if(tries==2) {
					/* Try sending power button, in case the box 
					   is turned off.  Need to wait for it to 
					   actually power on, and flush buffers
					   since it will probably send garbage. */
					debug("Sending KEY_POWER\n");
					if(send_keypress(KEY_POWER)>=0) {
						struct timespec sleeptime;
						sleeptime.tv_sec = POWER_SLEEP / 1000;
						sleeptime.tv_nsec = (POWER_SLEEP % 1000) * 1000 * 1000;
						while ((nanosleep(&sleeptime, &sleeptime) < 0) && (errno == EINTR));
					}
					serial_flush(TIMEOUT_TINY);
				}
#endif
				continue;
			}
			cur = newcur;

			if(cur != chan) {
				debug("channel is wrong; didn't switch?\n");
				continue;
			}

			debug("Success!\n");
			printf("%d\n",cur);
			return 0;
		}
		printf("%d\n",cur);
		return 0;
	}

	fprintf(stderr,"Communication failed after %d tries\n",tries);
	printf("%d\n",cur);
	return 1;
}
