/* Process data from the DCT2000 and fill data structures accordingly.
 * 
 * Based highly on DCTTest by Don Starr
 * http://www.starrsoft.com/freeware/DCTTest.zip
 *
 * Heavily modifed by Dave Johnson <dave-dctremote@davej.org>
 */ 

#include "serial.h"
#include "crc.h"
#include "debug.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/poll.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

static int serial_fd = -1;

/* Open and initialize serial port.
   Returns 0 on success, or -1 if error */
int serial_init(const char *port)
{
	struct termios t;

	serial_fd=open(port,O_RDWR);
	if(serial_fd<0)
		return -1;
	if(tcgetattr(serial_fd,&t)<0) {
		serial_close();
		return -1;
	}
	t.c_cflag |= CLOCAL;
	if(tcsetattr(serial_fd,TCSANOW,&t)<0) {
		serial_close();
		return -1;
	}
	
	bzero(&t,sizeof(t));
	t.c_iflag=IGNBRK|IGNPAR;
	t.c_cflag=CS8|CREAD|CLOCAL;
	t.c_cc[VMIN]=1;
	if(cfsetispeed(&t,B9600)<0) {
		serial_close();
		return -1;
	}
	if(tcsetattr(serial_fd,TCSANOW,&t)<0) {
		serial_close();
		return -1;
	}

	return 0;
}

/* Close a serial port */
void serial_close(void)
{
	close(serial_fd);
	serial_fd = -1;
}

/* returns -1 on timeout/error, otherwise 0 */
static int serial_getnextbyte(char *ch, int timeout_msec)
{
	struct pollfd pfd;

	pfd.fd = serial_fd;
	pfd.events = POLLIN|POLLERR;

	/* Wait 2 seconds for the byte */
	if(poll(&pfd, 1, timeout_msec)!=1)
		return -1;
	if(pfd.revents & POLLERR) {
		debug("POLLER\n");
		return -1;
	}
	
	if(read(serial_fd, ch, 1)!=1) {
		debug("READ: %s\n", strerror(errno));
		return -1;
	}

	return 0;
}

/* Read a packet from the serial port and return it.
   Return 0 on success, or -1 on error/timeout */
int serial_getpacket(_sPacket *p, int timeout_msec)
{
	_sProcessorInfo *pi = ProcessorInit();
	char ch;

	p->cpPayload=NULL;

	if(pi==NULL)
		return -1;

 more:
	if(serial_getnextbyte(&ch,timeout_msec)<0) {
		/* timeout */
		return -1;
	}
	if(ProcessByte(pi, ch)!=1)
		goto more;

	memcpy(p,&pi->packet,sizeof(_sPacket));
	free(pi);
	
	packet_received(p);

	return 0;
}

void serial_flush(int timeout_msec)
{
	char ch;
	while (serial_getnextbyte(&ch, timeout_msec)==0);
}

static void serial_write(char c)
{
	int ret;
	ret = write(serial_fd,&c,1);
	if (ret != 1)
		debug("WRITE: 0x%02x ret %d %s",
		      c, ret, ret < 0 ? strerror(errno): "");
}

static void serial_escape_write(char c)
{
	if(c==0x10) serial_write(c);
	serial_write(c);
}
	
/* Send a packet. */
void serial_sendpacket(_sPacket *p)
{
	int i;

	serial_write(0x10);
	serial_write(p->cPacketType);
	serial_escape_write(p->iLen >> 8);
	serial_escape_write(p->iLen & 0xff);
	serial_escape_write(p->cSeq);
	serial_escape_write(p->cID);
	for(i=0;i<p->iLen-2;i++)
		serial_escape_write(p->cpPayload[i]);
	serial_escape_write(p->iCRC >> 8);
	serial_escape_write(p->iCRC & 0xff);
	serial_write(0x10);
	serial_write(0x03);

	packet_sent(p);
}
